use serde::Deserialize;

#[derive(Debug, Deserialize, PartialEq)]
pub struct Dwml {
    pub head: Head,
    pub data: (Forecast, CurrentObservations),
}

impl std::str::FromStr for Dwml {
    type Err = quick_xml::DeError;

    fn from_str(s: &str) -> Result<Self, quick_xml::DeError> {
        quick_xml::de::from_str(s)
    }
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Head {
    pub product: Product,
    pub source: Source,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Forecast {
    pub location: Location,
    #[serde(rename = "time-layout")]
    pub time_layout: (TimeLayout, TimeLayout, TimeLayout),
    pub parameters: Parameters,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Location {
    #[serde(rename = "area-description")]
    pub area_description: Option<String>,
    pub description: Option<String>,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct TimeLayout {
    #[serde(rename = "@time-coordinate")]
    pub time_coordinate: TimeCoordinate,

    #[serde(rename = "layout-key")]
    pub layout_key: LayoutKey,

    #[serde(rename = "start-valid-time")]
    pub start_valid_time: Vec<StartValidTime>,
}

#[derive(Debug, Deserialize, PartialEq)]
pub enum TimeCoordinate {
    #[serde(rename = "local")]
    Local,
}

#[derive(Debug, Deserialize, PartialEq)]
pub enum LayoutKey {
    #[serde(rename = "k-p12h-n14-1")]
    K12Hourly,
    #[serde(rename = "k-p24h-n7-1")]
    K24Hourly1,
    #[serde(rename = "k-p24h-n7-2")]
    K24Hourly2,
    #[serde(rename = "k-p1h-n1-1")]
    K1Hour,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct StartValidTime {
    #[serde(rename = "@period-name")]
    pub period_name: String,

    #[serde(rename = "$value")]
    pub value: String,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct CurrentObservations {
    pub location: Location,
    #[serde(rename = "time-layout")]
    pub time_layout: TimeLayout,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Product {
    #[serde(rename = "creation-date")]
    pub creation_date: CreationDate,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct CreationDate {
    #[serde(rename = "@refresh-frequency")]
    pub refresh_frequency: String,

    #[serde(rename = "$value")]
    pub value: String,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Source {
    #[serde(rename = "production-center")]
    pub production_center: String,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Parameters {
    pub temperature: (Temperature, Temperature),

    #[serde(rename = "probability-of-precipitation")]
    pub probability_of_precipitation: ProbabilityOfPrecipitation,

    pub weather: Weather,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Temperature {
    #[serde(rename = "@type")]
    pub temp_type: TempType,

    #[serde(rename = "@units")]
    pub units: TempUnits,

    #[serde(rename = "@time-layout")]
    pub time_layout: LayoutKey,

    pub value: Vec<i32>,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct ProbabilityOfPrecipitation {
    #[serde(rename = "@time-layout")]
    pub time_layout: LayoutKey,

    pub value: Vec<PrecipitationValue>,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct Weather {
    #[serde(rename = "@time-layout")]
    pub time_layout: LayoutKey,

    #[serde(rename = "weather-conditions")]
    pub weather_conditions: Vec<WeatherCondition>,
}

#[derive(Debug, Deserialize, PartialEq)]
pub struct WeatherCondition {
    #[serde(rename = "@weather-summary")]
    pub weather_summary: String,
}

// If only there was a percentage that could convey the concept of
// no precipitation.
#[derive(Debug, Deserialize, PartialEq)]
pub struct PrecipitationValue {
    #[serde(rename = "$text")]
    pub value: Option<u32>,
}

#[derive(Debug, Deserialize, PartialEq)]
pub enum TempType {
    #[serde(rename = "apparent")]
    Apparent,
    #[serde(rename = "dew point")]
    DewPoint,
    #[serde(rename = "minimum")]
    Minimum,
    #[serde(rename = "maximum")]
    Maximum,
}

#[derive(Debug, Deserialize, PartialEq)]
pub enum TempUnits {
    Celsius,
    Fahrenheit,
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn parse() {
        let s = include_str!("../data/empire_city.xml");
        let _: Dwml = quick_xml::de::from_str(s).unwrap();

        let s = include_str!("../data/sunset_city.xml");
        let _: Dwml = quick_xml::de::from_str(s).unwrap();

        let s = include_str!("../data/windy_city.xml");
        let _: Dwml = quick_xml::de::from_str(s).unwrap();
    }
}
